import logging
import sys
from logging.config import dictConfig

# configure logging
dictConfig(
    {
        "version": 1,
        "formatters": {
            "json": {"()": "jamsuite.logger.JSONFormatter", "pretty": True},
        },
        "handlers": {
            "json_handler": {
                "class": "logging.StreamHandler",
                "stream": sys.stdout,
                "formatter": "json",
            },
        },
        "loggers": {
            "root": {
                "level": "INFO",
                "handlers": ["json_handler"],
            },
        },
    }
)

# let's log!
log = logging.getLogger(__name__)
log.info({"message": "Hello Mom!", "kid": "charm"})



import inspect
import json
import logging
from pathlib import Path

import pytest

from jamsuite.logger import DictMixin, ErrorMixin, JSONSerializer


class TestErrorFormatter(JSONSerializer, ErrorMixin, DictMixin, logging.Formatter):
    __test__ = False


@pytest.fixture
def log_message():
    return "Message here"


@pytest.fixture
def log_exception():
    return Exception("Rosie Tuesday")


@pytest.fixture
def log(logger_factory):
    yield logger_factory(formatter=TestErrorFormatter())


def test_exc_info_no_message(log, log_exception):
    """
    Steps:

    1 - log err_info with no message

    Result:

    * Raise TypeError from logging package
    """
    with pytest.raises(TypeError):
        log.error(exc_info=log_exception)


def test_exc_info_and_message(log, log_buffer, log_message, log_exception):
    """
    Steps:

    1 - log exc_info and message

    Result:

    * Both message and exception are present in log output
    * Exception without traceback has not traceback in the log output
    """
    log.error(log_message, exc_info=log_exception)
    log_entry = json.loads(log_buffer.getvalue())
    assert log_entry["message"] == log_message
    assert log_entry["log_exception"] == f"Exception({log_exception})"
    assert "log_traceback" not in log_entry


def test_exc_info_and_traceback(log, log_buffer, log_message, log_exception):
    """
    Steps:

    1 - raise exception
    2 - log it with message

    Result:

    * Both message and exception are present in log output
    * Exception has traceback with the one line from the test
    """
    try:
        raise log_exception
    except Exception as e:
        log.error(log_message, exc_info=e)

    log_entry = json.loads(log_buffer.getvalue())
    assert log_entry["message"] == log_message
    assert log_entry["log_exception"] == f"Exception({log_exception})"
    test_func = inspect.stack()[0].function
    test_path = str(Path(__file__).relative_to(Path.cwd()))
    assert len(log_entry["log_traceback"]) == 1
    assert test_func in log_entry["log_traceback"][0]
    assert test_path in log_entry["log_traceback"][0]



import importlib


def load_class_from_string(path: str):
    """
    Dynamically loads a class object from a fully qualified string path (e.g., 'module.sub.ClassName').

    Args:
        path (str): The string path to the class.

    Returns:
        The class object (not an instance).

    Raises:
        ImportError: If the module or class cannot be found.
    """
    try:
        # 1. Split the path into module and class name
        module_name, class_name = path.rsplit(".", 1)

        # 2. Import the module
        module = importlib.import_module(module_name)

        # 3. Get the class from the module
        return getattr(module, class_name)

    except (ValueError, ImportError, AttributeError) as e:
        # Handle cases where the path is malformed, module not found,
        # or class not found within the module
        raise ImportError(
            f"Could not load class '{path}'. Check if the module and class exist and are spelled correctly. Original error: {e}"
        )



import logging

from . import colors
from .mixins import DictMixin, ErrorMixin, MetadataMixin, PydanticMixin
from .serializers import JSONSerializer


class ColorFormatter(logging.Formatter):
    """
    Logging formatter supporting colorized output to the console

    You can customize the colors by setting `color_codes` in the Formatter constructor
    mapping log levels to ASCII color characters

    Note color codes include the ASCII control characters.
    See https://jafrog.com/2013/11/23/colors-in-terminal.html for more info.

    ```
    import logging

    ColorFormatter(color_codes={
        logging.DEBUG: '\x1b[32;3m',
        logging.INFO: '\x1b[34m',
        logging.WARNING: '\x1b[33;3m',
        logging.ERROR: '\x1b[31;1m',
        logging.CRITICAL: '\x1b[31;3m',
    })
    ```

    You can also change the ASCII reset character if you need to through the "reset_code" kwarg.
    """

    def __init__(self, *args, **kwargs):
        self.COLOR_CODES = kwargs.pop("color_codes", colors.LOG_COLORS)
        self.RESET_CODE = kwargs.pop("reset_code", colors.COLOR_RESET)
        super().__init__(*args, **kwargs)

    def format(self, record):
        if record.levelno in self.COLOR_CODES:
            record.color_on = self.COLOR_CODES[record.levelno]
            record.color_off = self.RESET_CODE
            return f"{record.color_on}{super().format(record)}{record.color_off}"
        return super().format(record)


class BaseFormatter(MetadataMixin, ErrorMixin, DictMixin, logging.Formatter):
    """
    Formats message as Dict and adds any error info (via DictFormatter and ErrorFormatter)

    Question: What if I want to customize how errors are added?

    Then you should extend ErrorFormatter (or make your own!) and build a new BaseFormatter
    for your use case.

    ex.
    ```
    class MyBaseFormatter(MetadataFormatter, MyErrorFormatter, DictFormatter):
        pass
    ```
    """

    pass


class JSONFormatter(JSONSerializer, PydanticMixin, BaseFormatter):
    pass


class ConsoleFormatter(ColorFormatter, JSONFormatter):
    """
    Logging formatter supporting colorized output and pretty JSON for the console
    """

    def __init__(self, *args, pretty=True, **kwargs):
        super().__init__(*args, pretty=pretty, **kwargs)



from .helpers import (
    RE_SENSITIVE_ATTR,
    format_dict,
    format_exception,
    format_template_debug,
    format_traceback_as_list,
    redact_secrets,
    set_nested_val,
)


class DictMixin:
    # From Docs: There are four keyword arguments in kwargs which are inspected: exc_info, stack_info, stacklevel and extra.
    # https://docs.python.org/3/library/logging.html#logging.Logger.debug

    def __init__(self, *args, **kwargs):
        self.message_key = kwargs.pop("message_key", "message")
        super().__init__(*args, **kwargs)

    def format(self, record) -> dict:
        """
        formats a records message into a dictionary

        does not handle exceptions or timestamps

        How would we expect a message like: log.info('good stuff', my_obj, error, extra=)
        to get logged?

        {
             "message": "good stuff",
             "args": [
                my_obj,
                error,
             ]
        }

        How would we expect a message like: log.info('good stuff', my_obj, error)
        to get logged?

        """
        return format_dict(record, message_key=self.message_key)


class ErrorMixin:
    def __init__(self, *args, **kwargs):
        self.exception_key = kwargs.pop("exception_key", "log_exception")
        self.template_debug_key = kwargs.pop("template_debug_key", "log_template_debug")
        self.traceback_key = kwargs.pop("traceback_key", "log_traceback")
        super().__init__(*args, **kwargs)

    def format(self, record) -> dict:
        """
        formats exception from record into msg dict

        you can log exceptions using exc_info, for example:

        ```
        try:
            ...
        catch Exception as e:
            log.error("an error occurred!", exc_info=e)
        ```

        You can also log exception as

        Right now, as-is, this will serialize as:

        Q. How do we add in support for record.stack_info (OR .sinfo)?  https://docs.python.org/3/library/logging.html#logrecord-attributes

        --> Needs more investigation
        """
        msg = super().format(record)

        if record.exc_info:
            self._add_exception_info(msg, record.exc_info)
        elif isinstance(msg["message"], Exception):
            # if message is an exception, log it as such
            self._add_exception_info(msg, msg["message"])
            del msg["message"]

        return msg

    def _add_exception_info(self, msg, exc):
        "convenience func to set exception values in message"
        set_nested_val(msg, self.exception_key, format_exception(exc))
        try:
            # for Jinja TemplateSyntaxErrors
            set_nested_val(msg, self.template_debug_key, format_template_debug(exc[1].template_debug))
        except (AttributeError, TypeError):
            pass
        if tb := format_traceback_as_list(exc):
            set_nested_val(msg, self.traceback_key, tb)


class TimestampMixin:
    """
    Add timestamp to the log message dictionary. By default adds the timestamp to `log_time` key in the dictionary.

    You can specifiy a different timestamp_key in the constructor to the Formatter:
    (ex. `TimestampFormatter(timestamp_key='@timestamp')` or `TimestampFormatter(timestamp_key='_meta.time')`).

    Uses standard Python Formatter.formatTime to format the time, per the `datefmt` setting

    See: https://docs.python.org/3/library/logging.html#logging.Formatter.formatTime for more info
    """

    def __init__(self, *args, **kwargs):
        self.timestamp_key = kwargs.pop("timestamp_key", "log_time")
        super().__init__(*args, **kwargs)

    def format(self, record) -> dict:
        msg = super().format(record)
        # add timestamp info
        record.asctime = self.formatTime(record, self.datefmt)
        set_nested_val(msg, self.timestamp_key, record.asctime)
        return msg


class MetadataMixin:
    """
    Add timestamp, level, and source to the log message dictionary.

    By default, adds:

    * timestamp to `log_time` key
    * source (record.name) to `log_source` key
    * level to `log_level` key

    You can specifiy a different keys timestamp_key in the constructor to the Formatter:

    For example `MetadataFormatter(timestamp_key='_time', source_key='log.logger', level_key='log.level')` will produce
    log messages like:

    ```
    {
        "_time": "2025-12-16T16:15:40.000Z",
        "log": {
            "logger": "my.module",
            "level": "INFO"
        },
    }
    ```

    Uses standard Python Formatter.formatTime to format the time, per the `datefmt` setting.
    See: https://docs.python.org/3/library/logging.html#logging.Formatter.formatTime for more info

    Question: How do I add more Metadata? For example I want record.lineno in my logfile. How do I add that?

    Simple, make your own formatter. It would look like this:

    ```python
    from jamsuite import logger

    # extend MetadataFormatter to add more metadata
    class MyMetadataFormatter(logger.MetadataFormatter):
        def format(self, record) -> dict:
            msg = super().format(record)
            # add your extra Metadata here
            msg["lineno"] = record.lineno
            return msg

    # Compose your MyJSONFormatter to use your extended Metadata
    class MyJSONFormatter(logger.JSONSerializer, MyMetadataFormatter, logger.ErrorFormatter, logger.DictFormatter):
        pass
    ```

    Then use MyJSONFormatter in your logging config.

    See [LogRecord.attributes documentation](https://docs.python.org/3/library/logging.html#logrecord-attributes)
    for what attributes are available.
    """

    def __init__(self, *args, **kwargs):
        self.timestamp_key = kwargs.pop("timestamp_key", "log_time")
        self.source_key = kwargs.pop("source_key", "log_source")
        self.level_key = kwargs.pop("level_key", "log_level")
        super().__init__(*args, **kwargs)

    def format(self, record) -> dict:
        msg = super().format(record)
        # add timestamp info
        record.asctime = self.formatTime(record, self.datefmt)
        set_nested_val(msg, self.timestamp_key, record.asctime)
        set_nested_val(msg, self.source_key, record.name)
        set_nested_val(msg, self.level_key, record.levelname)
        return msg


try:
    from pydantic import BaseModel as PydanticBaseModel

    HAVE_PYDANTIC = True
except ImportError:
    HAVE_PYDANTIC = False


class PydanticMixin:
    """
    if any first level value in message is a Pydantic model, serializes it using `model_dump(mode='json')`
    """

    def format(self, record) -> dict:
        if not HAVE_PYDANTIC:
            return super().format(record)

        msg = super().format(record)
        # don't mutate a dict while iterating over it
        msg_keys = list(msg.keys())
        for k in msg_keys:
            v = msg[k]
            if isinstance(v, PydanticBaseModel):
                msg[k] = v.model_dump(mode="json")
        return msg


class SecretMaskMixin:
    """
    Masks secret values in the log message by matching dictionary keys to a regex
    """

    def __init__(self, *args, **kwargs):
        """
        Params:

        - secrets_regex (regex) defaults to look for
          keys that contain the word SECRET, PASSWORD, or KEY, ignoring case
        """
        self.secrets_regex = kwargs.pop("secrets_regex", RE_SENSITIVE_ATTR)
        super().__init__(*args, **kwargs)

    def format(self, record):
        msg = super().format(record)
        redact_secrets(record.msg, regex=self.secrets_regex)
        return msg



from .colors import COLOR_BLUE, COLOR_BOLD_RED, COLOR_GREEN, COLOR_GREY, COLOR_RED, COLOR_RESET, COLOR_YELLOW, LOG_COLORS

# from .json_formatter import JSONFormatter, SensitiveJSONFormatter
from .formatters import BaseFormatter, ColorFormatter, ConsoleFormatter, JSONFormatter
from .helpers import (
    format_exception,
    format_template_debug,
    format_traceback_as_list,
    format_traceback_nopytest,
    get_log_level,
    get_logger,
)
from .mixins import DictMixin, ErrorMixin, MetadataMixin, PydanticMixin, TimestampMixin
from .serializers import JSONSerializer

__all__ = [
    COLOR_BLUE,
    COLOR_BOLD_RED,
    COLOR_GREEN,
    COLOR_GREY,
    COLOR_RED,
    COLOR_RESET,
    COLOR_YELLOW,
    LOG_COLORS,
    format_exception,
    format_template_debug,
    format_traceback_as_list,
    format_traceback_nopytest,
    get_logger,
    get_log_level,
    JSONSerializer,
    BaseFormatter,
    DictMixin,
    ErrorMixin,
    MetadataMixin,
    PydanticMixin,
    TimestampMixin,
    ColorFormatter,
    JSONFormatter,
    ConsoleFormatter,
]



import sys
from logging.config import dictConfig

from pydantic import BaseModel

from jamsuite.logger import get_logger


class A(BaseModel):
    a: str
    b: int
    c: dict[str, "A"]


dictConfig(
    {
        "version": 1,
        "formatters": {
            "json": {"()": "jamsuite.logger.JSONFormatter", "pretty": True},
        },
        "handlers": {
            "json_handler": {
                "class": "logging.StreamHandler",
                "stream": sys.stdout,
                "formatter": "json",
            },
        },
        "loggers": {
            "root": {
                "level": "INFO",
                "handlers": ["json_handler"],
            },
        },
    }
)

log = get_logger(__name__)
log.info(A(a="Hey", b=3, c={"Foo": A(a="There", b=4, c={})}))
log.info(A(a="Hey", b=3, c={"Foo": A(a="There", b=4, c={})}))



import datetime
import json
import logging


def default_converter(obj):
    "converter used by json.dumps to format datetime"
    if isinstance(obj, datetime.datetime):
        return obj.isoformat()
    return str(obj)


class JSONSerializer(logging.Formatter):
    def __init__(self, *args, **kwargs):
        self.pretty_json = kwargs.pop("pretty", False)
        super().__init__(*args, **kwargs)

    def format(self, record) -> str:
        msg = super().format(record)
        # serialize msg as JSON
        return json.dumps(msg, default=default_converter)



import logging
import os
import re
import traceback
from collections.abc import MutableMapping
from pathlib import Path

###
#
#   Convenience Funcs
#
###


def get_log_level(level: str | int) -> int:
    """
    set a log level either by number (ex. logging.WARNING) or string (ex. "DEBUG")

    raises ValueError if string value not found in logging.getLevelNamesMapping()
    """
    if isinstance(level, str):
        # convert str to int
        level_arg = level
        if not (level := logging.getLevelNamesMapping().get(level_arg)):
            raise ValueError(f"No log level found for {level_arg!r}")
    return level


def get_logger(name):
    """
    wrapper for logging.getLogger

    Usage:

        ```
        from jamsuite.logger import get_logger
        logger = get_logger(__name__)
        ```
    """
    return logging.getLogger(name)


def format_exception(e):
    """
    Given an exception, or tuple from exc_info, returns a string representation

    If the string representation is multiple lines, returns each line in a list
    """
    # if e is a tuple from exc_info, grab the exception
    # see https://docs.python.org/3/library/sys.html#sys.exc_info

    if isinstance(e, tuple):
        e = e[1]

    # stree = f"{type(e).__name__}({str(e)})"
    # convert exception to string, split into individual lines, remove empty lines
    lines = list(filter(None, str(e).split("\n")))

    # if only one line, return it as string
    if len(lines) == 1:
        return f"{type(e).__name__}({lines[0]})"
    return [type(e).__name__] + lines


# keep this as str instead of PosixPath for format_traceback_as_list()
SRCROOT = os.getenv("SRCROOT", str(Path.cwd()))


def format_traceback_as_list(e=None, filters=None, skip=0) -> list[str]:
    """
    returns a formatted traceback

    - remove leading path from filename for brevity
    - strip extra whitespace from the _original_line

    Args:
        e - an exception, or tuple from exc_info
        filters - any lines that start with these will be removed
        skip - don't include last ## lines
    """
    l = []  # noqa: E741
    if e:
        stack = traceback.extract_tb(e[2] if isinstance(e, tuple) else e.__traceback__)
    else:
        # don't include the call to format_traceback_as_list()
        skip = skip + 1
        stack = traceback.extract_stack()

    for frame_summary in stack:
        filename = frame_summary.filename
        # make filename more concise
        site_packages = filename.find("site-packages")
        if site_packages >= 0:
            # remove everything before site-packages, if present
            filename = filename[site_packages + 14 :]
        elif filename.startswith(SRCROOT):
            # remove everything through src folder
            filename = filename[len(SRCROOT) + 1 :]
        if filters and any(filename.startswith(f) for f in filters):
            continue
        original_lines = getattr(frame_summary, "_original_lines", getattr(frame_summary, "_original_line", ""))
        l.append(f"{filename} line {frame_summary.lineno} in {frame_summary.name}: {original_lines.strip()}")
    return l[:-skip] if skip else l


def format_traceback_nopytest(e=None, filters=["pytest", "_pytest", "pluggy"]):
    """filters out pytest from traceback"""
    return format_traceback_as_list(e=e, filters=filters, skip=1)


def format_template_debug(d):
    """
    Given a template_debug dict (d) from Jinja TemplateError, returns

    - remove leading path from filename for brevity
    - strip extra whitespace from the _original_line
    """
    return f"{d['name'][len(SRCROOT) + 1 :]} line {d['line']} in {d['during'].strip()}"


def format_multiline(s: str) -> str | list[str]:
    """
    splits multi-line string into a list of strings
    if string is not multiline, returns string as-is
    """
    m = s.split("\n")
    return m[0] if len(m) == 1 else m


def format_dict(record: logging.LogRecord, message_key="message") -> dict:
    """
    Formats log message as dict
    and sets the message to record.message
    also returning the message

    Formatting logic:
        + if record.msg is empty, sets empty dict {}
        + if record.msg is already a MutableMapping, sets it as-is
        + otherwise embeds record.message as "message" in the dict
        + If msg is a multiline string, converts it to a list

    Does not do anything with exceptions or timestamps
    """
    try:
        # if we already have record.message as dict return it
        if record.message and isinstance(record.message, MutableMapping):
            return record.message
    except AttributeError:
        pass

    # empty message return empty dict
    if not record.msg:
        m = {}

    # if message is already a MutableMapping (dict), return it
    if isinstance(record.msg, MutableMapping):
        m = record.msg
    else:
        # embed string message in dict, break multiline string into list
        m = {message_key: format_multiline(record.getMessage())}

    # TODO: Handle extra and defaults

    # keep the formatted message in the LogRecord
    record.message = m
    return m


def set_nested_val(d, path, value, delimiter="."):
    """
    Sets a value in a nested dictionary using a dot-separated string path.
    """
    keys = path.split(delimiter)

    # Iterate over all keys except the last one
    for key in keys[:-1]:
        # specific check to ensure we don't try to treat a non-dict as a dict
        # if an intermediate key exists but isn't a dict, we overwrite it
        if key not in d or not isinstance(d[key], dict):
            d[key] = {}
        d = d[key]

    # Set the value for the last key
    d[keys[-1]] = value


RE_URL_WITH_PASSWORD = re.compile(r'("[^"]*//[^/"]*@)([^:/"]*)([^"]*")')
RE_SENSITIVE_ATTR = re.compile(r"SECRET|PASSWORD|KEY", re.IGNORECASE)


def redact_secrets(d: MutableMapping, _level=1, regex=RE_SENSITIVE_ATTR, max_depth=1):
    """
    recurse through `d` up to `MAX_DEPTH` levels deep and replace
    values w/ "********" for any keys which match regex

    arguments:
    - `d` - log message dictionary

    kwarg params:
     - `regex` - regex to match against dictionary keys.
       (default=`RE_SENSITIVE_ATTR`, keys that contain the word SECRET, PASSWORD, or KEY, ignoring case)
     - `max_depth` how many levels to traverse the logging dictionary
    """
    if _level > max_depth:
        return
    for k, v in d.items():
        if RE_SENSITIVE_ATTR.match(k):
            d[k] = "********"
        elif isinstance(v, MutableMapping):
            redact_secrets(d, _level + 1, regex=regex, max_depth=max_depth)



import logging

# colors for pretty output!
# https://stackoverflow.com/a/56944256
# also https://chrisyeh96.github.io/2020/03/28/terminal-colors.html
COLOR_GREY = "\x1b[38;20m"
COLOR_GREEN = "\x1b[32;20m"
COLOR_BLUE = "\x1b[94;20m"
COLOR_YELLOW = "\x1b[33;20m"
COLOR_RED = "\x1b[31;20m"
COLOR_BOLD_RED = "\x1b[31;1m"
COLOR_RESET = "\x1b[0m"

LOG_COLORS = {
    logging.DEBUG: COLOR_GREEN,
    logging.INFO: COLOR_BLUE,
    logging.WARNING: COLOR_YELLOW,
    logging.ERROR: COLOR_RED,
    logging.CRITICAL: COLOR_BOLD_RED,
}



import logging
import uuid
from io import StringIO
from typing import Optional

import pytest

from jamsuite.logger import get_log_level

from .utils import load_class_from_string


@pytest.fixture
def log_buffer() -> StringIO:
    """Provides a StringIO stream to capture log output."""
    return StringIO()


@pytest.fixture
def logger_factory(log_buffer):
    created_handlers = []
    created_loggers = []

    def create_logger(name: Optional[str] = None, formatter: Optional[str | logging.Formatter] = None, level: str | int = logging.DEBUG):
        """
        Create an isolated logger with a StringIO handler for testing.
        """
        # Generate unique name to avoid conflicts
        if not name:
            name = f"test_logger_{uuid.uuid4().hex[:8]}"

        level = get_log_level(level)
        # create formatter
        if isinstance(formatter, str):
            FormatterClass = load_class_from_string(formatter)  # noqa: N806
            formatter = FormatterClass()

        # create handler from log_buffer
        handler = logging.StreamHandler(log_buffer)
        handler.setFormatter(formatter)
        handler.setLevel(level)
        created_handlers.append(handler)

        logger = logging.getLogger(name)
        logger.addHandler(handler)
        logger.setLevel(level)
        logger.propagate = False
        created_loggers.append(name)

        return logger

    yield create_logger

    # Cleanup: Remove all handlers and loggers
    for handler in created_handlers:
        handler.close()

    for logger_name in created_loggers:
        logger = logging.getLogger(logger_name)
        logger.handlers.clear()
        logger.filters.clear()
        # Remove from logging manager
        if logger_name in logging.Logger.manager.loggerDict:
            del logging.Logger.manager.loggerDict[logger_name]

    log_buffer.close()
